import sys


def compat_compose_dictionary(dictionary, **kwargs):
    """
    This method allows us the one line dictionary composition that is offered by the ** dictionary unpacking
    available in 3.6.

    :param dictionary: the dictionary to add the kwargs elements to.
    :param kwargs: a set of key/value pairs to add to `dictionary`.
    :return: the composed dictionary.
    """
    composed_dict = dictionary.copy()
    composed_dict.update(kwargs)

    return composed_dict


def compat_json(data, ignore_dicts=False):
    """

    :param data: Json Data we want to ensure compatibility on.
    :param ignore_dicts: should only be set to true when first called.
    :return: Python compatible 2.7 byte-strings when encountering unicode.
    """
    # if this is a unicode string, return its string representation
    if isinstance(data, unicode):
        return data.encode('utf-8')
    # if this is a list of values, return list of byte-string values
    if isinstance(data, list):
        return [compat_json(item, ignore_dicts=True) for item in data]
    # if this is a dictionary, return dictionary of byte-string keys and values
    # but only if we haven't already byte-string it
    if isinstance(data, dict) and not ignore_dicts:
        return {
            compat_json(key, ignore_dicts=True): compat_json(value, ignore_dicts=True)
            for key, value in data.iteritems()
        }
    # if it's anything else, return it in its original form
    return data

def compat_bytes(item, encoding=None):
    """
    This method is required because Python 2.7 `bytes` is simply an alias for `str`. Without this method,
    code execution would look something like:

    class clazz(object):

        def __bytes__(self):
            return bytes(5)


    Python 2.7:

    c = clazz()
    bytes(c)
    >>'<__main__.clazz object at 0x105171a90>'

    In this example, when `bytes(c)` is invoked, the interpreter then calls `str(c)`, and prints the above string.
    the method `__bytes__` is never invoked.

    Python 3.6:
    c = clazz()
    bytes(c)
    >>b'\x00\x00\x00\x00\x00'

    This is the expected and necessary behavior across both platforms.

    w/ compat_bytes method, we will ensure that the correct bytes method is always invoked, avoiding the `str` alias in
    2.7.

    :param item: this is the object who's bytes method needs to be invoked
    :param encoding: optional encoding parameter to handle the Python 3.6 two argument 'bytes' method.
    :return: a bytes object that functions the same across 3.6 and 2.7
    """
    if hasattr(item, '__bytes__'):
        return item.__bytes__()
    else:
        if encoding:
            return bytes(item, encoding)
        else:
            return bytes(item)


def compat_chr(item):
    """
    This is necessary to maintain compatibility across Python 2.7 and 3.6.
    In 3.6, 'chr' handles any unicode character, whereas in 2.7, `chr` only handles
    ASCII characters. Thankfully, the Python 2.7 method `unichr` provides the same
    functionality as 3.6 `chr`.

    :param item: a length 1 string who's `chr` method needs to be invoked
    :return: the unichr code point of the single character string, item
    """
    if sys.version >= '3.0':
        return chr(item)
    else:
        return unichr(item)



import hashlib
import struct
import sys
from binascii import hexlify, unhexlify

try:
    from Cryptodome.Cipher import AES
except ImportError:
    try:
        from Crypto.Cipher import AES
    except ImportError:
        raise ImportError("Missing dependency: pyCryptodome")

from .base58 import base58encode, base58decode
from .account import PrivateKey, PublicKey
from .utils import compat_bytes

default_prefix = "SHARE"

class PubKeyMismatchException(Exception):
    """Pubkeys do not match expected key"""


def get_shared_secret(priv, pub):
    """Derive the share secret between ``priv`` and ``pub``

    :param `Base58` priv: Private Key
    :param `Base58` pub: Public Key
    :return: Shared secret
    :rtype: hex

    The shared secret is generated such that::

        Pub(Alice) * Priv(Bob) = Pub(Bob) * Priv(Alice)

    """
    pub_point = pub.point()
    priv_point = int(repr(priv), 16)
    res = pub_point * priv_point
    res_hex = '%032x' % res.x()
    # Zero padding
    res_hex = '0' * (64 - len(res_hex)) + res_hex
    return hashlib.sha512(unhexlify(res_hex)).hexdigest()


def init_aes(shared_secret, nonce):
    """Initialize AES instance

    :param hex shared_secret: Shared Secret to use as encryption key
    :param int nonce: Random nonce
    :return: AES instance and checksum of the encryption key
    :rtype: length 2 tuple

    """
    " Seed "
    ss = unhexlify(shared_secret)
    n = struct.pack("<Q", int(nonce))
    encryption_key = hashlib.sha512(n + ss).hexdigest()
    " Check'sum' "
    check = hashlib.sha256(unhexlify(encryption_key)).digest()
    check = struct.unpack_from("<I", check[:4])[0]
    " AES "
    key = unhexlify(encryption_key[0:64])
    iv = unhexlify(encryption_key[64:96])
    return AES.new(key, AES.MODE_CBC, iv), check


def _pad(s, BS):
    numBytes = (BS - len(s) % BS)
    return s + numBytes * struct.pack('B', numBytes)


def _unpad(s, BS):
    count = int(struct.unpack('B', bytes(s[-1], 'ascii'))[0])
    if bytes(s[-count::], 'ascii') == count * struct.pack('B', count):
        return s[:-count]
    return s


def encode_memo(priv, pub, nonce, message, prefix="STM", **kwargs):
    """Encode a message with a shared secret between Alice and Bob

    :param PrivateKey priv: Private Key (of Alice)
    :param PublicKey pub: Public Key (of Bob)
    :param int nonce: Random nonce
    :param str message: Memo message
    :return: Encrypted message
    :rtype: hex

    """
    from .operations import Memo

    shared_secret = get_shared_secret(priv, pub)
    aes, check = init_aes(shared_secret, nonce)
    raw = bytes(message, 'utf8')
    " Padding "
    BS = 16
    if len(raw) % BS:
        raw = _pad(raw, BS)
    " Encryption "
    cipher = hexlify(aes.encrypt(raw)).decode("ascii")
    prefix = kwargs.pop("prefix", default_prefix)
    s = {
        "from": format(priv.pubkey, prefix),
        "to": format(pub, prefix),
        "nonce": nonce,
        "check": check,
        "encrypted": cipher,
        "from_priv": repr(priv),
        "to_pub": repr(pub),
        "shared_secret": shared_secret,
    }
    tx = Memo(**s, prefix=prefix)
    return "#" + base58encode(hexlify(bytes(tx)).decode("ascii"))


def decode_memo(priv, message, prefix="STM", pubkey=None, **kwargs):
    """Decode a message with a shared secret between Alice and Bob

    :param PrivateKey priv: Private Key (of Bob)
    :param base58encoded message: Encrypted Memo message
    :param str prefix: Used prefix for pubkey presentation
    :param PublicKey pubkey: (Optional) pubkey to validate message against
    :return: Decrypted message
    :rtype: str
    :raise ValueError: if message cannot be decoded as valid UTF-8
           string

    """
    " decode structure "
    raw = base58decode(message[1:])
    from_key = PublicKey(raw[:66], prefix=prefix)
    raw = raw[66:]
    to_key = PublicKey(raw[:66], prefix=prefix)
    raw = raw[66:]
    nonce = str(struct.unpack_from("<Q", unhexlify(raw[:16]))[0])
    raw = raw[16:]
    check = struct.unpack_from("<I", unhexlify(raw[:8]))[0]
    raw = raw[8:]
    cipher = raw

    if repr(to_key) == repr(priv.pubkey):
        shared_secret = get_shared_secret(priv, from_key)
    elif repr(from_key) == repr(priv.pubkey):
        shared_secret = get_shared_secret(priv, to_key)
    else:
        raise ValueError("Incorrect PrivateKey")

    if pubkey and get_shared_secret(priv, pubkey) != shared_secret:
        raise PubKeyMismatchException("Pubkeys do not match expected key")


    " Init encryption "
    aes, checksum = init_aes(shared_secret, nonce)

    " Check "
    assert check == checksum, "Checksum failure"

    " Encryption "
    # remove the varint prefix (FIXME, long messages!)
    message = cipher[2:]
    message = aes.decrypt(unhexlify(bytes(message, 'ascii')))
    try:
        return _unpad(message.decode('utf8'), 16)
    except:
        raise ValueError(message)


def involved_keys(message):
    " decode structure "
    raw = base58decode(message[1:])
    from_key = PublicKey(raw[:66])
    raw = raw[66:]
    to_key = PublicKey(raw[:66])

    return [from_key, to_key]



# coding=utf-8
import logging

from funcy.seqs import first
from hivebase.chains import known_chains
from hivebase.http_client import HttpClient
from hivebase.storage import configStorage
from hivebase.transactions import SignedTransaction
from hivebase.types import PointInTime

from .block import Block
from .blockchain import Blockchain
from .post import Post
from .utils import resolve_identifier
from .utils import compat_compose_dictionary
from .instance import get_config_node_list

logger = logging.getLogger(__name__)


class Hived(HttpClient):
    """ Connect to the Hive network.

        Args:

            nodes (list): A list of Hive HTTP RPC nodes to connect to. If
            not provided, official hive.blog nodes will be used.

        Returns:

            Hived class instance. It can be used to execute commands
            against Hive node.

        Example:

           If you would like to override the official hive.blog nodes
           (default), you can pass your own.  When currently used node goes
           offline, ``Hived`` will automatically fail-over to the next
           available node.

           .. code-block:: python

               nodes = [
                   'https://hived.yournode1.com',
                   'https://hived.yournode2.com',
               ]

               h = Hived(nodes)

       """

    def __init__(self, nodes=None, **kwargs):
        if not nodes:
            nodes = get_config_node_list() or ["https://api.hive.blog"]

        super(Hived, self).__init__(nodes, **kwargs)

    @property
    def chain_params(self):
        """ Identify the connected network. This call returns a
            dictionary with keys chain_id, prefix, and other chain
            specific settings
        """
        props = self.get_config()
        chain_id = props["HIVE_CHAIN_ID"]

        chain = None
        for supported_chain in known_chains.values():
            if chain_id == supported_chain["chain_id"]:
                chain = supported_chain

        if chain is None:
            raise Exception("The chain you are connecting " + "to is not supported")

        return chain

    def get_replies(self, author, skip_own=True):
        """ Get replies for an author

            :param str author: Show replies for this author
            :param bool skip_own: Do not show my own replies
        """
        state = self.get_state("/@%s/recent-replies" % author)
        replies = state["accounts"][author].get("recent_replies", [])
        discussions = []
        for reply in replies:
            post = state["content"][reply]
            if skip_own and post["author"] == author:
                continue
            discussions.append(Post(post, hived_instance=self))
        return discussions

    def get_promoted(self):
        """ Get promoted posts
        """
        state = self.get_state("/promoted")
        # why is there a empty key in the struct?
        promoted = state["discussion_idx"][""].get("promoted", [])
        r = []
        for p in promoted:
            post = state["content"].get(p)
            r.append(Post(post, hived_instance=self))
        return r

    def get_posts(self, limit=10, sort="hot", category=None, start=None):
        """ Get multiple posts in an array.

            :param int limit: Limit the list of posts by ``limit``
            :param str sort: Sort the list by "recent" or "payout"
            :param str category: Only show posts in this category
            :param str start: Show posts after this post. Takes an
                              identifier of the form ``author/permlink``
        """

        discussion_query = {
            "tag": category,
            "limit": limit,
        }
        if start:
            author, permlink = resolve_identifier(start)
            discussion_query["start_author"] = author
            discussion_query["start_permlink"] = permlink

        if sort not in [
            "trending",
            "created",
            "active",
            "cashout",
            "payout",
            "votes",
            "children",
            "hot",
        ]:
            raise Exception("Invalid choice of '--sort'!")

        func = getattr(self, "get_discussions_by_%s" % sort)
        r = []
        for p in func(discussion_query):
            r.append(Post(p, hived_instance=self))
        return r

    def stream_comments(self, *args, **kwargs):
        """ Generator that yields posts when they come in

            To be used in a for loop that returns an instance of `Post()`.
        """
        for c in Blockchain(
            mode=kwargs.get("mode", "irreversible"), hived_instance=self,
        ).stream("comment", *args, **kwargs):
            yield Post(c, hived_instance=self)

    @property
    def last_irreversible_block_num(self):
        """ Newest irreversible block number. """
        return self.get_dynamic_global_properties()["last_irreversible_block_num"]

    @property
    def head_block_number(self):
        """ Newest block number. """
        return self.get_dynamic_global_properties()["head_block_number"]

    def get_account(self, account):
        """ Lookup account information such as user profile, public keys,
        balances, etc.

        Args:
            account (str): Hive username that we are looking up.

        Returns:
            dict: Account information.

        """
        return first(self.call("get_accounts", [account]))

    def get_all_usernames(self, last_user=""):
        """ Fetch the full list of Hive usernames. """
        usernames = self.lookup_accounts(last_user, 1000)
        batch = []
        while len(batch) != 1:
            batch = self.lookup_accounts(usernames[-1], 1000)
            usernames += batch[1:]

        return usernames

    def _get_blocks(self, blocks):
        """ Fetch multiple blocks from hived at once.

        Warning: This method does not ensure that all blocks are returned,
        or that the results are ordered.  You will probably want to use
        `hived.get_blocks()` instead.

        Args:
            blocks (list): A list, or a set of block numbers.

        Returns:
            A generator with results.

        """
        results = self.call_multi_with_futures("get_block", blocks, max_workers=10)
        return (
            compat_compose_dictionary(x, block_num=int(x["block_id"][:8], base=16))
            for x in results
            if x
        )

    def get_blocks(self, block_nums):
        """ Fetch multiple blocks from hived at once, given a range.

        Args:

            block_nums (list): A list of all block numbers we would like to
            tech.

        Returns:

            dict: An ensured and ordered list of all `get_block` results.

        """
        required = set(block_nums)
        available = set()
        missing = required - available
        blocks = {}

        while missing:
            for block in self._get_blocks(missing):
                blocks[block["block_num"]] = block

            available = set(blocks.keys())
            missing = required - available

        return [blocks[x] for x in block_nums]

    def get_blocks_range(self, start, end):
        """ Fetch multiple blocks from hived at once, given a range.

        Args:

            start (int): The number of the block to start with

            end (int): The number of the block at the end of the range. Not
            included in results.

        Returns:
            dict: An ensured and ordered list of all `get_block` results.

        """
        return self.get_blocks(list(range(start, end)))

    def get_reward_fund(self, fund_name="post"):
        """ Get details for a reward fund.

        Right now the only pool available is 'post'.

        Example:

            .. code-block:: python

                s.get_reward_fund('post')

            ::

                {'content_constant': '2000000000000',
                 'id': 0,
                 'last_update': '2017-04-09T19:18:57',
                 'name': 'post',
                 'percent_content_rewards': 10000,
                 'percent_curation_rewards': 2500,
                 'recent_claims': '10971122501158586840771928156084',
                 'reward_balance': '555660.895 HIVE'}

        """
        return self.call("get_reward_fund", fund_name, api="database_api")

    def get_expiring_vesting_delegations(self, account, start, limit):
        """ get_expiring_vesting_delegations """
        return self.call(
            "get_expiring_vesting_delegations",
            account,
            start,
            limit,
            api="database_api",
        )

    def get_trending_tags(self, after_tag, limit):
        """ get_trending_tags """
        return self.call("get_trending_tags", after_tag, limit, api="database_api")

    def get_tags_used_by_author(self, account):
        """ get_tags_used_by_author """
        return self.call("get_tags_used_by_author", account, api="database_api")

    def get_discussions_by_trending(self, discussion_query):
        """ get_discussions_by_trending """
        return self.call(
            "get_discussions_by_trending", discussion_query, api="database_api"
        )

    def get_comment_discussions_by_payout(self, discussion_query):
        """ get_comment_discussions_by_payout """
        return self.call(
            "get_comment_discussions_by_payout", discussion_query, api="database_api"
        )

    def get_post_discussions_by_payout(self, discussion_query):
        """ get_post_discussions_by_payout """
        return self.call(
            "get_post_discussions_by_payout", discussion_query, api="database_api"
        )

    def get_discussions_by_created(self, discussion_query):
        """ get_discussions_by_created """
        return self.call(
            "get_discussions_by_created", discussion_query, api="database_api"
        )

    def get_discussions_by_active(self, discussion_query):
        """ get_discussions_by_active """
        return self.call(
            "get_discussions_by_active", discussion_query, api="database_api"
        )

    def get_discussions_by_cashout(self, discussion_query):
        """ get_discussions_by_cashout """
        return self.call(
            "get_discussions_by_cashout", discussion_query, api="database_api"
        )

    def get_discussions_by_payout(self, discussion_query):
        """ get_discussions_by_payout """
        return self.call(
            "get_discussions_by_payout", discussion_query, api="database_api"
        )

    def get_discussions_by_votes(self, discussion_query):
        """ get_discussions_by_votes """
        return self.call(
            "get_discussions_by_votes", discussion_query, api="database_api"
        )

    def get_discussions_by_children(self, discussion_query):
        """ get_discussions_by_children """
        return self.call(
            "get_discussions_by_children", discussion_query, api="database_api"
        )

    def get_discussions_by_hot(self, discussion_query):
        """ get_discussions_by_hot """
        return self.call("get_discussions_by_hot", discussion_query, api="database_api")

    def get_discussions_by_feed(self, discussion_query):
        """ get_discussions_by_feed """
        return self.call(
            "get_discussions_by_feed", discussion_query, api="database_api"
        )

    def get_discussions_by_blog(self, discussion_query):
        """ get_discussions_by_blog """
        return self.call(
            "get_discussions_by_blog", discussion_query, api="database_api"
        )

    def get_discussions_by_comments(self, discussion_query):
        """ get_discussions_by_comments """
        return self.call(
            "get_discussions_by_comments", discussion_query, api="database_api"
        )

    def get_discussions_by_promoted(self, discussion_query):
        """ get_discussions_by_promoted """
        return self.call(
            "get_discussions_by_promoted", discussion_query, api="database_api"
        )

    def get_block_header(self, block_num):
        """ Get block headers, given a block number.

        Args:
           block_num (int) number.

        Returns:
           dict headers in a JSON compatible format.

        Example:

            .. code-block:: python

               s.get_block_headers(8888888)

            ::

                {'extensions': [],
                 'previous': '0087a2372163ff5c5838b09589ce281d5a564f66',
                 'timestamp': '2017-01-29T02:47:33',

                 'transaction_merkle_root':
                 '4ddc419e531cccee6da660057d606d11aab9f3a5',
                 'witness': 'chainsquad.com'}
        """
        return self.call("get_block_header", block_num, api="database_api")

    def get_block(self, block_num):
        """ Get the full block, transactions and all, given a block number.

        Args:
            block_num (int) number.

        Returns:
            dict in a JSON compatible format.

        """
        return self.call("get_block", block_num, api="database_api")

    def get_ops_in_block(self, block_num, virtual_only):
        """ get_ops_in_block """
        return self.call(
            "get_ops_in_block", block_num, virtual_only, api="database_api"
        )

    def get_state(self, path):
        """ get_state """
        return self.call("get_state", path, api="database_api")

    def get_config(self):
        """ Get internal chain configuration. """
        return self.call("get_config", api="database_api")

    def get_dynamic_global_properties(self):
        """ get_dynamic_global_properties """
        return self.call("get_dynamic_global_properties", api="database_api")

    def get_chain_properties(self):
        """ Get witness elected chain properties.

        ::

            {'account_creation_fee': '30.000 HIVE',
             'maximum_block_size': 65536,
             'hbd_interest_rate': 250}

        """
        return self.call("get_chain_properties", api="database_api")

    def get_feed_history(self):
        """ Get the hourly averages of witness reported HIVE/HBD prices.

        ::

            {'current_median_history':
                {'base': '0.093 HBD', 'quote': '1.010 HIVE'},
             'id': 0,
             'price_history': [{'base': '0.092 HBD', 'quote': '1.010 HIVE'},
              {'base': '0.093 HBD', 'quote': '1.020 HIVE'},
              {'base': '0.093 HBD', 'quote': '1.010 HIVE'},
              {'base': '0.094 HBD', 'quote': '1.020 HIVE'},
              {'base': '0.093 HBD', 'quote': '1.010 HIVE'},

        """
        return self.call("get_feed_history", api="database_api")

    def get_current_median_history_price(self):
        """ Get the average HIVE/HBD price.

        This price is based on moving average of witness reported price feeds.

        ::

            {'base': '0.093 HBD', 'quote': '1.010 HIVE'}

        """
        return self.call("get_current_median_history_price", api="database_api")

    def get_witness_schedule(self):
        """ get_witness_schedule """
        return self.call("get_witness_schedule", api="database_api")

    def get_hardfork_version(self):
        """ Get the current version of the chain.

        Note:
            This is not the same as latest minor version.

        """
        return self.call("get_hardfork_version", api="database_api")

    def get_next_scheduled_hardfork(self):
        """ get_next_scheduled_hardfork """
        return self.call("get_next_scheduled_hardfork", api="database_api")

    def get_accounts(self, account_names):
        """ Lookup account information such as user profile, public keys,
        balances, etc.

        This method is same as ``get_account``, but supports querying for
        multiple accounts at the time.

        """
        return self.call("get_accounts", account_names, api="database_api")

    def get_account_references(self, account_id):
        """ get_account_references """
        return self.call("get_account_references", account_id, api="database_api")

    def lookup_account_names(self, account_names):
        """ lookup_account_names """
        return self.call("lookup_account_names", account_names, api="database_api")

    def lookup_accounts(self, after, limit):
        """Get a list of usernames from all registered accounts.

        Args:

            after (str, int): Username to start with. If '', 0 or -1, it
            will start at beginning.

            limit (int): How many results to return.

        Returns:
            list: List of usernames in requested chunk.

        """
        return self.call("lookup_accounts", after, limit, api="database_api")

    def get_account_count(self):
        """ How many accounts are currently registered on Hive? """
        return self.call("get_account_count", api="database_api")

    def get_conversion_requests(self, account):
        """ get_conversion_requests """
        return self.call("get_conversion_requests", account, api="database_api")

    def get_account_history(self, account, index_from, limit):
        """ History of all operations for a given account.

        Args:

           account (str): Hive username that we are looking up.

           index_from (int): The highest database index we take as a
           starting point.

           limit (int): How many items are we interested in.

        Returns:

           list: List of operations.

        Example:

           To get the latest (newest) operations from a given user
           ``furion``, we should set the ``index_from`` to -1.  This is the
           same as saying `give me the highest index there is`.

           .. code-block :: python

              s.get_account_history('furion', index_from=-1, limit=3)

           This will yield 3 recent operations like so:

           ::

              [[69974,
                {'block': 9941972,
                 'op': ['vote',
                  {'author': 'breezin',
                   'permlink': 'raising-childr....',
                   'voter': 'furion',
                   'weight': 900}],
                 'op_in_trx': 0,
                 'timestamp': '2017-03-06T17:09:48',
                 'trx_id': '87f9176faccc7096b5ffb5d12bfdb41b3c0b2955',
                 'trx_in_block': 5,
                 'virtual_op': 0}],
               [69975,
                {'block': 9942005,
                 'op': ['curation_reward',
                  {'comment_author': 'leongkhan',
                   'comment_permlink': 'steem-investor-report-5-march-2017',
                   'curator': 'furion',
                   'reward': '112.397602 VESTS'}],
                 'op_in_trx': 1,
                 'timestamp': '2017-03-06T17:11:30',
                 'trx_id': '0000000000000000000000000000000000000000',
                 'trx_in_block': 5,
                 'virtual_op': 0}],
               [69976,
                {'block': 9942006,
                 'op': ['vote',
                  {'author': 'ejhaasteem',
                   'permlink': 'life-of-fishermen-in-aceh',
                   'voter': 'furion',
                   'weight': 100}],
                 'op_in_trx': 0,
                 'timestamp': '2017-03-06T17:11:30',
                 'trx_id': '955018ac8efe298bd90b45a4fbd15b9df7e00be4',
                 'trx_in_block': 7,
                 'virtual_op': 0}]]

           If we want to query for a particular range of indexes, we need
           to consider both `index_from` and `limit` fields.  Remember,
           `index_from` works backwards, so if we set it to 100, we will
           get items `100, 99, 98, 97...`.

           For example, if we'd like to get the first 100 operations the
           user did, we would write:

           .. code-block:: python

              s.get_account_history('furion', index_from=100, limit=100)

           We can get the next 100 items by running:

           .. code-block:: python

              s.get_account_history('furion', index_from=200, limit=100)


        """
        return self.call(
            "get_account_history", account, index_from, limit, api="database_api"
        )

    def get_owner_history(self, account):
        """ get_owner_history """
        return self.call("get_owner_history", account, api="database_api")

    def get_recovery_request(self, account):
        """ get_recovery_request """
        return self.call("get_recovery_request", account, api="database_api")

    def get_escrow(self, from_account, escrow_id):
        """ get_escrow """
        return self.call("get_escrow", from_account, escrow_id, api="database_api")

    def get_withdraw_routes(self, account, withdraw_route_type):
        """ get_withdraw_routes """
        return self.call(
            "get_withdraw_routes", account, withdraw_route_type, api="database_api"
        )

    def get_account_bandwidth(self, account, bandwidth_type):
        """ get_account_bandwidth """
        return self.call(
            "get_account_bandwidth", account, bandwidth_type, api="database_api"
        )

    def get_order_book(self, limit):
        """ Get the internal market order book.

        This method will return both bids and asks.

        Args:
            limit (int): How many levels deep into the book to show.

        Returns:
            dict: Order book.

        Example:

            .. code-block:: python

               s.get_order_book(2)

            Outputs:

            ::

                {'asks': [{'created': '2017-03-06T21:29:54',
                   'order_price':
                   {'base': '513.571 HIVE', 'quote': '50.000 HBD'},
                   'real_price': '0.09735752213423265',
                   'hbd': 50000,
                   'hive': 513571},
                  {'created': '2017-03-06T21:01:39',
                   'order_price':
                   {'base': '63.288 HIVE', 'quote': '6.204 HBD'},
                   'real_price': '0.09802806219188472',
                   'hbd': 6204,
                   'hive': 63288}],
                 'bids': [{'created': '2017-03-06T21:29:51',
                   'order_price':
                   {'base': '50.000 HBD', 'quote': '516.503 HIVE'},
                   'real_price': '0.09680485882947436',
                   'hbd': 50000,
                   'hive': 516503},
                  {'created': '2017-03-06T17:30:24',
                   'order_price':
                   {'base': '36.385 HBD', 'quote': '379.608 HIVE'},
                   'real_price': '0.09584887568228277',
                   'hbd': 36385,
                   'hive': 379608}]}


        """
        return self.call("get_order_book", limit, api="database_api")

    def get_open_orders(self, account):
        """ get_open_orders """
        return self.call("get_open_orders", account, api="database_api")

    def get_liquidity_queue(self, start_account, limit):
        """ Get the liquidity queue.

        Warning:

            This feature is currently not in use, and might be deprecated
            in the future.

        """
        return self.call(
            "get_liquidity_queue", start_account, limit, api="database_api"
        )

    def get_transaction_hex(self, signed_transaction):
        """ get_transaction_hex """
        return self.call("get_transaction_hex", signed_transaction, api="database_api")

    def get_transaction(self, transaction_id):
        """ get_transaction """
        return self.call("get_transaction", transaction_id, api="database_api")

    def get_required_signatures(self, signed_transaction, available_keys):
        """ get_required_signatures """
        return self.call(
            "get_required_signatures",
            signed_transaction,
            available_keys,
            api="database_api",
        )

    def get_potential_signatures(self, signed_transaction):
        """ get_potential_signatures """
        return self.call(
            "get_potential_signatures", signed_transaction, api="database_api"
        )

    def verify_authority(self, signed_transaction):
        """ verify_authority """
        return self.call("verify_authority", signed_transaction, api="database_api")

    def verify_account_authority(self, account, keys):
        """ verify_account_authority """
        return self.call("verify_account_authority", account, keys, api="database_api")

    def get_active_votes(self, author, permlink):
        """ Get all votes for the given post.

        Args:
            author (str): OP's Hive username.

            permlink (str): Post identifier following the username. It
            looks like slug-ified title.

        Returns:
            list: List of votes.

        Example:
            .. code-block:: python

               s.get_active_votes('mynameisbrian',
               'steemifying-idioms-there-s-no-use-crying-over-spilt-milk')

            Output:

            ::

               [{'percent': 10000,
                 'reputation': '36418980678',
                 'rshares': 356981288,
                 'time': '2017-03-06T20:04:18',
                 'voter': 'dailystuff',
                 'weight': '2287202760855'},
                 ...
                {'percent': 10000,
                 'reputation': 3386400109,
                 'rshares': 364252169,
                 'time': '2017-03-06T19:32:45',
                 'voter': 'flourish',
                 'weight': '2334690471157'}]
        """
        return self.call("get_active_votes", author, permlink, api="database_api")

    def get_account_votes(self, account):
        """ All votes the given account ever made.

        Returned votes are in the following format:
        ::

           {'authorperm':
           'alwaysfelicia/time-line-of-best-time...',
           'percent': 100, 'rshares': 709227399, 'time':
           '2016-08-07T16:06:24', 'weight': '3241351576115042'},


        Args:
            account (str): Hive username that we are looking up.

        Returns:
            list: List of votes.


        """
        return self.call("get_account_votes", account, api="database_api")

    def get_content(self, author, permlink):
        """ get_content """
        return self.call("get_content", author, permlink, api="database_api")

    def get_content_replies(self, author, permlink):
        """ get_content_replies """
        return self.call("get_content_replies", author, permlink, api="database_api")

    def get_discussions_by_author_before_date(
        self, author, start_permlink, before_date, limit
    ):
        """ get_discussions_by_author_before_date """
        return self.call(
            "get_discussions_by_author_before_date",
            author,
            start_permlink,
            before_date,
            limit,
            api="database_api",
        )

    def get_replies_by_last_update(self, account, start_permlink, limit):
        """ get_replies_by_last_update """
        return self.call(
            "get_replies_by_last_update",
            account,
            start_permlink,
            limit,
            api="database_api",
        )

    def get_witnesses(self, witness_ids):
        """ get_witnesses """
        return self.call("get_witnesses", witness_ids, api="database_api")

    def get_witness_by_account(self, account):
        """ get_witness_by_account """
        return self.call("get_witness_by_account", account, api="database_api")

    def get_witnesses_by_vote(self, from_account, limit):
        """ get_witnesses_by_vote """
        return self.call(
            "get_witnesses_by_vote", from_account, limit, api="database_api"
        )

    def lookup_witness_accounts(self, from_account, limit):
        """ lookup_witness_accounts """
        return self.call(
            "lookup_witness_accounts", from_account, limit, api="database_api"
        )

    def get_witness_count(self):
        """ get_witness_count """
        return self.call("get_witness_count", api="database_api")

    def get_active_witnesses(self):
        """ Get a list of currently active witnesses. """
        return self.call("get_active_witnesses", api="database_api")

    def get_vesting_delegations(self, account, from_account, limit):
        """ get_vesting_delegations """
        return self.call(
            "get_vesting_delegations", account, from_account, limit, api="database_api"
        )

    def login(self, username, password):
        """ login """
        return self.call("login", username, password, api="login_api")

    def get_api_by_name(self, api_name):
        """ get_api_by_name """
        return self.call("get_api_by_name", api_name, api="login_api")

    def get_version(self):
        """ Get hived version of the node currently connected to. """
        return self.call("get_version", api="database_api")

    def get_followers(self, account, start_follower, follow_type, limit):
        """ get_followers """
        return self.call(
            "get_followers",
            account,
            start_follower,
            follow_type,
            limit,
            api="follow_api",
        )

    def get_following(self, account, start_follower, follow_type, limit):
        """ get_following """
        return self.call(
            "get_following",
            account,
            start_follower,
            follow_type,
            limit,
            api="follow_api",
        )

    def get_follow_count(self, account):
        """ get_follow_count """
        return self.call("get_follow_count", account, api="follow_api")

    def get_feed_entries(self, account, entry_id, limit):
        """ get_feed_entries """
        return self.call("get_feed_entries", account, entry_id, limit, api="follow_api")

    def get_feed(self, account, entry_id, limit):
        """ get_feed """
        return self.call("get_feed", account, entry_id, limit, api="follow_api")

    def get_blog_entries(self, account, entry_id, limit):
        """ get_blog_entries """
        return self.call("get_blog_entries", account, entry_id, limit, api="follow_api")

    def get_blog(self, account, entry_id, limit):
        """ get_blog """
        return self.call("get_blog", account, entry_id, limit, api="follow_api")

    def get_account_reputations(self, account, limit):
        """ get_account_reputations """
        return self.call("get_account_reputations", account, limit, api="follow_api")

    def get_reblogged_by(self, author, permlink):
        """ get_reblogged_by """
        return self.call("get_reblogged_by", author, permlink, api="follow_api")

    def get_blog_authors(self, blog_account):
        """ get_blog_authors """
        return self.call("get_blog_authors", blog_account, api="follow_api")

    def broadcast_transaction(self, signed_transaction):
        """ broadcast_transaction """
        return self.call(
            "broadcast_transaction", signed_transaction, api="network_broadcast_api"
        )

    def broadcast_transaction_synchronous(self, signed_transaction):
        """ broadcast_transaction_synchronous """
        return self.call(
            "broadcast_transaction_synchronous",
            signed_transaction,
            api="network_broadcast_api",
        )

    def broadcast_block(self, block):
        """ broadcast_block """
        return self.call("broadcast_block", block, api="network_broadcast_api")

    def set_max_block_age(self, max_block_age):
        """ set_max_block_age """
        return self.call(
            "set_max_block_age", max_block_age, api="network_broadcast_api"
        )

    def get_ticker(self):
        """ Returns the market ticker for the internal HBD:HIVE market. """
        return self.call("get_ticker", api="market_history_api")

    def get_volume(self):
        """ Returns the market volume for the past 24 hours. """
        return self.call("get_volume", api="market_history_api")

    def get_trade_history(self, start, end, limit):
        """ Returns the trade history for the internal HBD:HIVE market. """
        return self.call(
            "get_trade_history", start, end, limit, api="market_history_api"
        )

    def get_recent_trades(self, limit):
        """ Returns the N most recent trades for the internal HBD:HIVE
        market. """

        return self.call("get_recent_trades", limit, api="market_history_api")

    def get_market_history(self, bucket_seconds, start, end):
        """ Returns the market history for the internal HBD:HIVE market. """
        return self.call(
            "get_market_history", bucket_seconds, start, end, api="market_history_api"
        )

    def get_market_history_buckets(self):
        """ Returns the bucket seconds being tracked by the plugin. """
        return self.call("get_market_history_buckets", api="market_history_api")

    def get_key_references(self, public_keys):
        """ get_key_references """
        if type(public_keys) == str:
            public_keys = [public_keys]
        return self.call("get_key_references", public_keys, api="account_by_key_api")


if __name__ == "__main__":
    h = Hived()
    print(h.get_account_count())



# coding=utf-8
import json
import logging
import socket
import time
import sys
from functools import partial
from itertools import cycle
import concurrent.futures
import certifi
import urllib3
from hivebase.exceptions import RPCError, RPCErrorRecoverable
from urllib3.connection import HTTPConnection
from urllib3.exceptions import MaxRetryError, ReadTimeoutError, ProtocolError

if sys.version >= "3.5":
    from http.client import RemoteDisconnected

if sys.version >= "3.0":
    from urllib.parse import urlparse
else:
    from urlparse import urlparse
    from httplib import HTTPException

logger = logging.getLogger(__name__)


class HttpClient(object):
    """Simple Hive JSON-HTTP-RPC API

    This class serves as an abstraction layer for easy use of the Hive API.

    Args:
      nodes (list): A list of Hive HTTP RPC nodes to connect to.

    .. code-block:: python

       from hive.http_client import HttpClient

       rpc = HttpClient(['https://hived-node1.com',
       'https://hived-node2.com'])

    any call available to that port can be issued using the instance
    via the syntax ``rpc.call('command', *parameters)``.

    Example:

    .. code-block:: python

       rpc.call(
           'get_followers',
           'furion', 'abit', 'blog', 10,
           api='follow_api'
       )

    """

    # set of endpoints which were detected to not support condenser_api
    non_appbase_nodes = set()

    def __init__(self, nodes, **kwargs):
        self.re_raise = kwargs.get("re_raise", True)
        self.max_workers = kwargs.get("max_workers", None)

        num_pools = kwargs.get("num_pools", 10)
        maxsize = kwargs.get("maxsize", 10)
        timeout = kwargs.get("timeout", 60)
        retries = kwargs.get("retries", 20)
        pool_block = kwargs.get("pool_block", False)
        tcp_keepalive = kwargs.get("tcp_keepalive", True)

        if tcp_keepalive:
            socket_options = HTTPConnection.default_socket_options + [
                (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
            ]
        else:
            socket_options = HTTPConnection.default_socket_options

        self.http = urllib3.poolmanager.PoolManager(
            num_pools=num_pools,
            maxsize=maxsize,
            block=pool_block,
            timeout=timeout,
            retries=retries,
            socket_options=socket_options,
            headers={"Content-Type": "application/json"},
            cert_reqs="CERT_REQUIRED",
            ca_certs=certifi.where(),
        )
        """
            urlopen(method, url, body=None, headers=None, retries=None,
            redirect=True, assert_same_host=True, timeout=<object object>,
            pool_timeout=None, release_conn=None, chunked=False, body_pos=None,
            **response_kw)
        """

        self.nodes = cycle(self.sanitize_nodes(nodes))
        self.url = ""
        self.request = None
        self.next_node()

        log_level = kwargs.get("log_level", logging.INFO)
        logger.setLevel(log_level)

    def _curr_node_downgraded(self):
        return self.url in HttpClient.non_appbase_nodes

    def _downgrade_curr_node(self):
        HttpClient.non_appbase_nodes.add(self.url)

    def _is_error_recoverable(self, error):
        assert "message" in error, "missing error msg key: {}".format(error)
        assert "code" in error, "missing error code key: {}".format(error)
        message = error["message"]
        code = error["code"]

        # common hived error
        # {"code"=>-32003, "message"=>"Unable to acquire database lock"}
        if message == "Unable to acquire database lock":
            return True

        # rare hived error
        # {"code"=>-32000, "message"=>"Unknown exception", "data"=>"0 exception: unspecified\nUnknown Exception\n[...]"}
        if message == "Unknown exception":
            return True

        # generic jussi error
        # {'code': -32603, 'message': 'Internal Error', 'data': {'error_id': 'c7a15140-f306-4727-acbd-b5e8f3717e9b',
        #         'request': {'amzn_trace_id': 'Root=1-5ad4cb9f-9bc86fbca98d9a180850fb80', 'jussi_request_id': None}}}
        if message == "Internal Error" and code == -32603:
            return True

        # jussi timeout error
        # {'code': 1100, 'message': 'Bad or missing upstream response',
        #  'data':{'error_id': 'fc3650d7-72...', 'exception': TimeoutError()}}
        if message == "Bad or missing upstream response" and code == 1100:
            return True

        return False

    def next_node(self):
        """Switch to the next available node.

        This method will change base URL of our requests.

        Use it when the current node goes down to change to a fallback
        node.

        """
        self.set_node(next(self.nodes))

    def set_node(self, node_url):
        """ Change current node to provided node URL. """
        self.url = node_url
        self.request = partial(self.http.urlopen, "POST", self.url)

    @property
    def hostname(self):
        return urlparse(self.url).hostname

    @staticmethod
    def json_rpc_body(name, *args, **kwargs):
        """Build request body for hived RPC requests.

        Args:

            name (str): Name of a method we are trying to call. (ie:
            `get_accounts`)

            args: A list of arguments belonging to the calling method.

            api (None, str): If api is provided (ie: `follow_api`),
             we generate a body that uses `call` method appropriately.

            as_json (bool): Should this function return json as dictionary
            or string.

            _id (int): This is an arbitrary number that can be used for
            request/response tracking in multi-threaded scenarios.

        Returns:

            (dict,str): If `as_json` is set to `True`, we get json
            formatted as a string.

            Otherwise, a Python dictionary is returned.

        """

        # if kwargs is non-empty after this, it becomes the call params
        as_json = kwargs.pop("as_json", True)
        api = kwargs.pop("api", None)
        _id = kwargs.pop("_id", 0)

        # `kwargs` for object-style param, `args` for list-style. pick one.
        assert not (kwargs and args), "fail - passed array AND object args"
        params = kwargs if kwargs else args

        if api and name == "get_version":
            api = "database_api"
            name = api + "." + name
            params = {}
        if api and name != api + ".get_version":
            body = {
                "jsonrpc": "2.0",
                "id": _id,
                "method": "call",
                "params": [api, name, params],
            }
        else:
            body = {"jsonrpc": "2.0", "id": _id, "method": name, "params": params}

        if as_json:
            return json.dumps(body, ensure_ascii=False).encode("utf8")

        return body

    def call(self, name, *args, **kwargs):
        """Call a remote procedure in hived.

        Warnings:

            This command will auto-retry in case of node failure, as well
            as handle node fail-over.

        """

        # tuple of Exceptions which are eligible for retry
        retry_exceptions = (
            MaxRetryError,
            ReadTimeoutError,
            ProtocolError,
            RPCErrorRecoverable,
        )

        if sys.version > "3.5":
            retry_exceptions += (
                json.decoder.JSONDecodeError,
                RemoteDisconnected,
            )
        else:
            retry_exceptions += (ValueError,)

        if sys.version > "3.0":
            retry_exceptions += (ConnectionResetError,)
        else:
            retry_exceptions += (HTTPException,)

        tries = 0
        while True:
            try:

                body_kwargs = kwargs.copy()
                if not self._curr_node_downgraded():
                    body_kwargs["api"] = "condenser_api"
                body = HttpClient.json_rpc_body(name, *args, **body_kwargs)
                response = self.request(body=body)
                success_codes = tuple(list(response.REDIRECT_STATUSES) + [200])
                if response.status not in success_codes:
                    raise RPCErrorRecoverable(
                        "non-200 response: %s from %s"
                        % (response.status, self.hostname)
                    )

                result = json.loads(response.data.decode("utf-8"))
                assert result, "result entirely blank"

                if "error" in result:
                    # legacy (pre-appbase) nodes always return err code 1
                    legacy = result["error"]["code"] == 1
                    detail = result["error"]["message"]

                    # some errors have no data key (db lock error)
                    if "data" not in result["error"]:
                        error = "error"
                    # some errors have no name key (jussi errors)
                    elif "name" not in result["error"]["data"]:
                        error = result["error"]
                        if "exception" in error["data"]:
                            error = error["data"]["exception"]
                        else:
                            error = "unspecified error"
                    else:
                        from pprint import pprint

                        if isinstance(result["error"]["data"], list):
                            error = result["error"]["data"]["name"]
                        else:
                            error = result["error"]["data"]

                    if legacy:
                        detail = ":".join(detail.split("\n")[0:2])
                        if not self._curr_node_downgraded():
                            self._downgrade_curr_node()
                            logging.error("Downgrade-retry %s", self.hostname)
                            continue

                    detail = "%s from %s (%s) in %s" % (
                        json.dumps(error, indent=4),
                        self.hostname,
                        detail,
                        name,
                    )

                    if self._is_error_recoverable(result["error"]):
                        raise RPCErrorRecoverable(detail)
                    else:
                        raise RPCError(detail)

                return result["result"]

            except retry_exceptions as e:
                if e == ValueError and "JSON" not in e.args[0]:
                    raise e  # (python<3.5 lacks json.decoder.JSONDecodeError)
                if tries >= 10:
                    logging.error(
                        "Failed after %d attempts -- %s: %s",
                        tries,
                        e.__class__.__name__,
                        e,
                    )
                    raise e
                tries += 1
                logging.warning(
                    "Retry in %ds -- %s: %s", tries, e.__class__.__name__, e
                )
                time.sleep(tries)
                self.next_node()
                continue

            # TODO: unclear why this case is here; need to explicitly
            #       define exceptions for which we refuse to retry.
            except Exception as e:
                extra = dict(err=e, request=self.request)
                logger.error(
                    "Unexpected exception! Please report at "
                    + "https://github.com/pharesim/hive-python/issues"
                    + " -- %s: %s",
                    e.__class__.__name__,
                    e,
                    extra=extra,
                )
                raise e

    def call_multi_with_futures(self, name, params, api=None, max_workers=None):
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            # Start the load operations and mark each future with its URL
            def ensure_list(val):
                return val if isinstance(val, (list, tuple, set)) else [val]

            futures = (
                executor.submit(self.call, name, *ensure_list(param), api=api)
                for param in params
            )
            for future in concurrent.futures.as_completed(futures):
                yield future.result()

    def sanitize_nodes(self, nodes):
        """

        This method is designed to explicitly validate the user defined
        nodes that are passed to http_client. If left unvalidated, improper
        input causes a variety of explicit-to-red herring errors in the code base.
        This will fail loudly in the event that incorrect input has been passed.

        There are three types of input allowed when defining nodes.
        1. a string of a single node url. ie nodes='https://api.hive.blog'
        2. a comma separated string of several node url's.
            nodes='https://api.hive.blog,<<community node>>,<<your personal node>>'
        3. a list of string node url's.
            nodes=['https://api.hive.blog','<<community node>>','<<your personal node>>']

        Any other input will result in a ValueError being thrown.

        :param nodes: the nodes argument passed to http_client
        :return: a list of node url's.
        """

        if self._isString(nodes):
            nodes = nodes.split(",")
        elif isinstance(nodes, list):
            if not all(self._isString(node) for node in nodes):
                raise ValueError("All nodes in list must be a string.")
        else:
            raise ValueError(
                "nodes arg must be a "
                "comma separated string of node url's, "
                "a single string url, "
                "or a list of strings."
            )

        return nodes

    def _isString(self, input):
        return isinstance(input, str) or (
            sys.version < "3.0" and isinstance(input, unicode)
        )



default_prefix = "SHARE"

known_chains = {
    "Shareberry-Testnet": {
        "chain_id": "a3ee3a34dd82dcd8e4f9ed14202c6d35734ea4c8d9c0e9daee48ec859d135b6c",
        "prefix": "SHARE",
        "steem_symbol": "SHTO",
        "sbd_symbol": "UNDEFINED",
        "vests_symbol": "SHPO",
    },
    "Shareberry": {
        "chain_id": "a5f5e08a1f98a474b3fd943f07fa69ae5f95307c8e94e5913647f627182e2b93",
        "prefix": "SHARE",
        "steem_symbol": "SHTO",
        "sbd_symbol": "UNDEFINED",
        "vests_symbol": "SHPO",
    },
}



import importlib
import json
from binascii import hexlify, unhexlify
import re
import struct
from collections import OrderedDict

from .utils import compat_bytes
from .account import PublicKey
from .operationids import operations
from .types import (
    Int16,
    Uint16,
    Uint32,
    Uint64,
    String,
    HexString,
    Bytes,
    Array,
    PointInTime,
    Bool,
    Optional,
    Map,
    Id,
    JsonObj,
    StaticVariant,
)

default_prefix = "STM"

asset_precision = {
    "STEEM": 3,
    "SHTO": 3,
    "HIVE": 3,
    "VESTS": 6,
    "SHPO": 6,
    "HBD": 3,
    "SBD": 3,
}


class Operation:
    def __init__(self, op):
        if isinstance(op, list) and len(op) == 2:
            if isinstance(op[0], int):
                self.opId = op[0]
                name = self.get_operation_name_for_id(self.opId)
            else:
                self.opId = operations.get(op[0], None)
                name = op[0]
                if self.opId is None:
                    raise ValueError("Unknown operation")

            # convert method name like feed_publish to class
            # name like FeedPublish
            self.name = self.to_class_name(name)
            try:
                klass = self.get_class(self.name)
            except:  # noqa FIXME(sneak)
                raise NotImplementedError("Unimplemented Operation %s" % self.name)
            else:
                self.op = klass(op[1])
        else:
            self.op = op
            # class name like FeedPublish
            self.name = type(self.op).__name__
            self.opId = operations[self.to_method_name(self.name)]

    @staticmethod
    def get_operation_name_for_id(_id):
        """Convert an operation id into the corresponding string"""
        for key, value in operations.items():
            if value == int(_id):
                return key

    @staticmethod
    def to_class_name(method_name):
        """Take a name of a method, like feed_publish and turn it into
        class name like FeedPublish."""
        return "".join(map(str.title, method_name.split("_")))

    @staticmethod
    def to_method_name(class_name):
        """Take a name of a class, like FeedPublish and turn it into
        method name like feed_publish."""
        words = re.findall("[A-Z][^A-Z]*", class_name)
        return "_".join(map(str.lower, words))

    @staticmethod
    def get_class(class_name):
        """ Given name of a class from `operations`, return real class. """
        module = importlib.import_module("hivebase.operations")
        return getattr(module, class_name)

    def __bytes__(self):
        return compat_bytes(Id(self.opId)) + compat_bytes(self.op)

    def __str__(self):
        return json.dumps([self.get_operation_name_for_id(self.opId), self.op.json()])


class GrapheneObject(object):
    """Core abstraction class

    This class is used for any JSON reflected object in Graphene.

    * ``instance.__json__()``: encodes data into json format
    * ``bytes(instance)``: encodes data into wire format
    * ``str(instances)``: dumps json object as string

    """

    def __init__(self, data=None):
        self.data = data

    def __bytes__(self):
        if self.data is None:
            return bytes()
        b = b""
        for name, value in self.data.items():
            if isinstance(value, str):
                b += compat_bytes(value, "utf-8")
            else:
                b += compat_bytes(value)
        return b

    def __json__(self):
        if self.data is None:
            return {}
        d = {}  # JSON output is *not* ordered
        for name, value in self.data.items():
            if isinstance(value, Optional) and value.isempty():
                continue

            if isinstance(value, String):
                d.update({name: str(value)})
            else:
                d.update({name: JsonObj(value)})
        return d

    def __str__(self):
        return json.dumps(self.__json__())

    def toJson(self):
        return self.__json__()

    def json(self):
        return self.__json__()


class Permission(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            prefix = kwargs.pop("prefix", default_prefix)

            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            # Sort keys (FIXME: ideally, the sorting is part of Public
            # Key and not located here)
            kwargs["key_auths"] = sorted(
                kwargs["key_auths"],
                key=lambda x: repr(PublicKey(x[0], prefix=prefix)),
                reverse=False,
            )
            kwargs["account_auths"] = sorted(
                kwargs["account_auths"],
                key=lambda x: x[0],
                reverse=False,
            )

            accountAuths = Map(
                [[String(e[0]), Uint16(e[1])] for e in kwargs["account_auths"]]
            )
            keyAuths = Map(
                [
                    [PublicKey(e[0], prefix=prefix), Uint16(e[1])]
                    for e in kwargs["key_auths"]
                ]
            )
            super(Permission, self).__init__(
                OrderedDict(
                    [
                        ("weight_threshold", Uint32(int(kwargs["weight_threshold"]))),
                        ("account_auths", accountAuths),
                        ("key_auths", keyAuths),
                    ]
                )
            )


class Memo(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            prefix = kwargs.pop("prefix", default_prefix)

            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            super(Memo, self).__init__(
                OrderedDict(
                    [
                        ("from", PublicKey(kwargs["from"], prefix=prefix)),
                        ("to", PublicKey(kwargs["to"], prefix=prefix)),
                        ("nonce", Uint64(int(kwargs["nonce"]))),
                        ("check", Uint32(int(kwargs["check"]))),
                        ("encrypted", Bytes(kwargs["encrypted"])),
                    ]
                )
            )


class Vote(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super(Vote, self).__init__(
                OrderedDict(
                    [
                        ("voter", String(kwargs["voter"])),
                        ("author", String(kwargs["author"])),
                        ("permlink", String(kwargs["permlink"])),
                        ("weight", Int16(kwargs["weight"])),
                    ]
                )
            )


class Comment(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            meta = ""
            if "json_metadata" in kwargs and kwargs["json_metadata"]:
                if isinstance(kwargs["json_metadata"], dict) or isinstance(
                    kwargs["json_metadata"], list
                ):
                    meta = json.dumps(kwargs["json_metadata"])
                else:
                    meta = kwargs["json_metadata"]

            super(Comment, self).__init__(
                OrderedDict(
                    [
                        ("parent_author", String(kwargs["parent_author"])),
                        ("parent_permlink", String(kwargs["parent_permlink"])),
                        ("author", String(kwargs["author"])),
                        ("permlink", String(kwargs["permlink"])),
                        ("title", String(kwargs["title"])),
                        ("body", String(kwargs["body"])),
                        ("json_metadata", String(meta)),
                    ]
                )
            )


class Amount:
    def __init__(self, d):
        self.amount, self.asset = d.strip().split(" ")
        self.amount = float(self.amount)

        if self.asset in asset_precision:
            self.precision = asset_precision[self.asset]
        else:
            raise Exception("Asset unknown")

    def __bytes__(self):
        # padding
        asset = self.asset + "\x00" * (7 - len(self.asset))
        amount = round(float(self.amount) * 10 ** self.precision)
        return (
            struct.pack("<q", amount)
            + struct.pack("<b", self.precision)
            + compat_bytes(asset, "ascii")
        )

    def __str__(self):
        return "{:.{}f} {}".format(self.amount, self.precision, self.asset)


class ExchangeRate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            super(ExchangeRate, self).__init__(
                OrderedDict(
                    [
                        ("base", Amount(kwargs["base"])),
                        ("quote", Amount(kwargs["quote"])),
                    ]
                )
            )


class WitnessProps(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            super(WitnessProps, self).__init__(
                OrderedDict(
                    [
                        (
                            "account_creation_fee",
                            Amount(kwargs["account_creation_fee"]),
                        ),
                        ("maximum_block_size", Uint32(kwargs["maximum_block_size"])),
                        ("hbd_interest_rate", Uint16(kwargs["hbd_interest_rate"])),
                    ]
                )
            )


class Beneficiary(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super(Beneficiary, self).__init__(
                OrderedDict(
                    [
                        ("account", String(kwargs["account"])),
                        ("weight", Int16(kwargs["weight"])),
                    ]
                )
            )


class Beneficiaries(GrapheneObject):
    def __init__(self, kwargs):
        super(Beneficiaries, self).__init__(
            OrderedDict(
                [
                    (
                        "beneficiaries",
                        Array([Beneficiary(o) for o in kwargs["beneficiaries"]]),
                    ),
                ]
            )
        )


class CommentOptionExtensions(StaticVariant):
    """Serialize Comment Payout Beneficiaries.

    Args:
        beneficiaries (list): A static_variant containing beneficiaries.

    Example:

        ::

            [0,
                {'beneficiaries': [
                    {'account': 'furion', 'weight': 10000}
                ]}
            ]
    """

    def __init__(self, o):
        type_id, data = o
        if type_id == 0:
            data = Beneficiaries(data)
        else:
            raise Exception("Unknown CommentOptionExtension")
        StaticVariant.__init__(self, data, type_id)


########################################################
# Actual Operations
########################################################


class AccountCreate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            prefix = kwargs.pop("prefix", default_prefix)

            assert (
                len(kwargs["new_account_name"]) <= 16
            ), "Account name must be at most 16 chars long"

            meta = ""
            if "json_metadata" in kwargs and kwargs["json_metadata"]:
                if isinstance(kwargs["json_metadata"], dict):
                    meta = json.dumps(kwargs["json_metadata"])
                else:
                    meta = kwargs["json_metadata"]
            super(AccountCreate, self).__init__(
                OrderedDict(
                    [
                        ("fee", Amount(kwargs["fee"])),
                        ("creator", String(kwargs["creator"])),
                        ("new_account_name", String(kwargs["new_account_name"])),
                        ("owner", Permission(kwargs["owner"], prefix=prefix)),
                        ("active", Permission(kwargs["active"], prefix=prefix)),
                        ("posting", Permission(kwargs["posting"], prefix=prefix)),
                        ("memo_key", PublicKey(kwargs["memo_key"], prefix=prefix)),
                        ("json_metadata", String(meta)),
                    ]
                )
            )


class AccountCreateWithDelegation(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            prefix = kwargs.pop("prefix", default_prefix)

            assert (
                len(kwargs["new_account_name"]) <= 16
            ), "Account name must be at most 16 chars long"

            meta = ""
            if "json_metadata" in kwargs and kwargs["json_metadata"]:
                if isinstance(kwargs["json_metadata"], dict):
                    meta = json.dumps(kwargs["json_metadata"])
                else:
                    meta = kwargs["json_metadata"]
            super(AccountCreateWithDelegation, self).__init__(
                OrderedDict(
                    [
                        ("fee", Amount(kwargs["fee"])),
                        ("delegation", Amount(kwargs["delegation"])),
                        ("creator", String(kwargs["creator"])),
                        ("new_account_name", String(kwargs["new_account_name"])),
                        ("owner", Permission(kwargs["owner"], prefix=prefix)),
                        ("active", Permission(kwargs["active"], prefix=prefix)),
                        ("posting", Permission(kwargs["posting"], prefix=prefix)),
                        ("memo_key", PublicKey(kwargs["memo_key"], prefix=prefix)),
                        ("json_metadata", String(meta)),
                        ("extensions", Array([])),
                    ]
                )
            )


class AccountUpdate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            prefix = kwargs.pop("prefix", default_prefix)

            meta = ""
            if "json_metadata" in kwargs and kwargs["json_metadata"]:
                if isinstance(kwargs["json_metadata"], dict):
                    meta = json.dumps(kwargs["json_metadata"])
                else:
                    meta = kwargs["json_metadata"]

            owner = (
                Permission(kwargs["owner"], prefix=prefix)
                if "owner" in kwargs
                else None
            )
            active = (
                Permission(kwargs["active"], prefix=prefix)
                if "active" in kwargs
                else None
            )
            posting = (
                Permission(kwargs["posting"], prefix=prefix)
                if "posting" in kwargs
                else None
            )

            super(AccountUpdate, self).__init__(
                OrderedDict(
                    [
                        ("account", String(kwargs["account"])),
                        ("owner", Optional(owner)),
                        ("active", Optional(active)),
                        ("posting", Optional(posting)),
                        ("memo_key", PublicKey(kwargs["memo_key"], prefix=prefix)),
                        ("json_metadata", String(meta)),
                    ]
                )
            )


class ChangeRecoveryAccount(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            super(ChangeRecoveryAccount, self).__init__(
                OrderedDict(
                    [
                        ("account_to_recover", String(kwargs["account_to_recover"])),
                        (
                            "new_recovery_account",
                            String(kwargs["new_recovery_account"]),
                        ),
                        ("extensions", Array([])),
                    ]
                )
            )


class Transfer(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            if "memo" not in kwargs:
                kwargs["memo"] = ""
            super(Transfer, self).__init__(
                OrderedDict(
                    [
                        ("from", String(kwargs["from"])),
                        ("to", String(kwargs["to"])),
                        ("amount", Amount(kwargs["amount"])),
                        ("memo", String(kwargs["memo"])),
                    ]
                )
            )


class TransferToVesting(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super(TransferToVesting, self).__init__(
                OrderedDict(
                    [
                        ("from", String(kwargs["from"])),
                        ("to", String(kwargs["to"])),
                        ("amount", Amount(kwargs["amount"])),
                    ]
                )
            )


class WithdrawVesting(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super(WithdrawVesting, self).__init__(
                OrderedDict(
                    [
                        ("account", String(kwargs["account"])),
                        ("vesting_shares", Amount(kwargs["vesting_shares"])),
                    ]
                )
            )


class ClaimRewardBalance(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super(ClaimRewardBalance, self).__init__(
                OrderedDict(
                    [
                        ("account", String(kwargs["account"])),
                        ("reward_hive", Amount(kwargs["reward_hive"])),
                        ("reward_hbd", Amount(kwargs["reward_hbd"])),
                        ("reward_vests", Amount(kwargs["reward_vests"])),
                    ]
                )
            )


class DelegateVestingShares(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super(DelegateVestingShares, self).__init__(
                OrderedDict(
                    [
                        ("delegator", String(kwargs["delegator"])),
                        ("delegatee", String(kwargs["delegatee"])),
                        ("vesting_shares", Amount(kwargs["vesting_shares"])),
                    ]
                )
            )


class LimitOrderCreate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super(LimitOrderCreate, self).__init__(
                OrderedDict(
                    [
                        ("owner", String(kwargs["owner"])),
                        ("orderid", Uint32(int(kwargs["orderid"]))),
                        ("amount_to_sell", Amount(kwargs["amount_to_sell"])),
                        ("min_to_receive", Amount(kwargs["min_to_receive"])),
                        ("fill_or_kill", Bool(kwargs["fill_or_kill"])),
                        ("expiration", PointInTime(kwargs["expiration"])),
                    ]
                )
            )


class LimitOrderCancel(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super(LimitOrderCancel, self).__init__(
                OrderedDict(
                    [
                        ("owner", String(kwargs["owner"])),
                        ("orderid", Uint32(int(kwargs["orderid"]))),
                    ]
                )
            )


class SetWithdrawVestingRoute(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super(SetWithdrawVestingRoute, self).__init__(
                OrderedDict(
                    [
                        ("from_account", String(kwargs["from_account"])),
                        ("to_account", String(kwargs["to_account"])),
                        ("percent", Uint16((kwargs["percent"]))),
                        ("auto_vest", Bool(kwargs["auto_vest"])),
                    ]
                )
            )


class Convert(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super(Convert, self).__init__(
                OrderedDict(
                    [
                        ("owner", String(kwargs["owner"])),
                        ("requestid", Uint32(kwargs["requestid"])),
                        ("amount", Amount(kwargs["amount"])),
                    ]
                )
            )


class FeedPublish(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super(FeedPublish, self).__init__(
                OrderedDict(
                    [
                        ("publisher", String(kwargs["publisher"])),
                        ("exchange_rate", ExchangeRate(kwargs["exchange_rate"])),
                    ]
                )
            )


class WitnessUpdate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            prefix = kwargs.pop("prefix", default_prefix)

            if not kwargs["block_signing_key"]:
                kwargs[
                    "block_signing_key"
                ] = "STM1111111111111111111111111111111114T1Anm"
            super(WitnessUpdate, self).__init__(
                OrderedDict(
                    [
                        ("owner", String(kwargs["owner"])),
                        ("url", String(kwargs["url"])),
                        (
                            "block_signing_key",
                            PublicKey(kwargs["block_signing_key"], prefix=prefix),
                        ),
                        ("props", WitnessProps(kwargs["props"])),
                        ("fee", Amount(kwargs["fee"])),
                    ]
                )
            )


class WitnessSetProperties(GrapheneObject):
    """
    Based on https://github.com/holgern/beem/blob/6cc303d1b0fdfb096da78d3ff331aaa79a18ad8f/beembase/operations.py#L278-L318
    """

    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            prefix = kwargs.pop("prefix", default_prefix)
            extensions = Array([])
            props = {}
            for k in kwargs["props"]:
                if "key" == k[0]:
                    block_signing_key = PublicKey(k[1], prefix=prefix)
                    props["key"] = repr(block_signing_key)
                elif "new_signing_key" == k[0]:
                    new_signing_key = PublicKey(k[1], prefix=prefix)
                    props["new_signing_key"] = repr(new_signing_key)
            for k in kwargs["props"]:
                if k[0] in ["key", "new_signing_key"]:
                    continue
                if isinstance(k[1], str):
                    is_hex = re.match(r"^[0-9a-fA-F]+$", k[1] or "") is not None
                else:
                    is_hex = False
                if isinstance(k[1], int) and k[0] in [
                    "account_subsidy_budget",
                    "account_subsidy_decay",
                    "maximum_block_size",
                    "hbd_interest_rate",
                ]:
                    props[k[0]] = (hexlify(Uint32(k[1]).__bytes__())).decode()
                elif not isinstance(k[1], str) and k[0] in ["account_creation_fee"]:
                    props[k[0]] = (hexlify(Amount(k[1]).__bytes__())).decode()
                elif (
                    not is_hex
                    and isinstance(k[1], str)
                    and k[0] in ["account_creation_fee"]
                ):
                    props[k[0]] = (hexlify(Amount(k[1]).__bytes__())).decode()
                elif not isinstance(k[1], str) and k[0] in ["hbd_exchange_rate"]:
                    props[k[0]] = (hexlify(ExchangeRate(k[1]).__bytes__())).decode()
                elif not is_hex and k[0] in ["url"]:
                    props[k[0]] = (hexlify(String(k[1]).__bytes__())).decode()
                else:
                    props[k[0]] = k[1]
            props_list = []
            for k in props:
                props_list.append(([String(k), HexString(props[k])]))
            props_list = sorted(
                props_list,
                key=lambda x: str(x[0]),
                reverse=False,
            )
            map_props = Map(props_list)

            super(WitnessSetProperties, self).__init__(
                OrderedDict(
                    [
                        ("owner", String(kwargs["owner"])),
                        ("props", map_props),
                        ("extensions", extensions),
                    ]
                )
            )


class AccountWitnessVote(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super(AccountWitnessVote, self).__init__(
                OrderedDict(
                    [
                        ("account", String(kwargs["account"])),
                        ("witness", String(kwargs["witness"])),
                        ("approve", Bool(bool(kwargs["approve"]))),
                    ]
                )
            )


class CustomJson(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            if "json" in kwargs and kwargs["json"]:
                if isinstance(kwargs["json"], dict) or isinstance(kwargs["json"], list):
                    js = json.dumps(kwargs["json"])
                else:
                    js = kwargs["json"]

            if len(kwargs["id"]) > 32:
                raise Exception("'id' too long")

            super(CustomJson, self).__init__(
                OrderedDict(
                    [
                        (
                            "required_auths",
                            Array([String(o) for o in kwargs["required_auths"]]),
                        ),
                        (
                            "required_posting_auths",
                            Array(
                                [String(o) for o in kwargs["required_posting_auths"]]
                            ),
                        ),
                        ("id", String(kwargs["id"])),
                        ("json", String(js)),
                    ]
                )
            )


class CommentOptions(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            # handle beneficiaries
            if "beneficiaries" in kwargs and kwargs["beneficiaries"]:
                kwargs["extensions"] = [[0, {"beneficiaries": kwargs["beneficiaries"]}]]

            extensions = Array([])
            if "extensions" in kwargs and kwargs["extensions"]:
                extensions = Array(
                    [CommentOptionExtensions(o) for o in kwargs["extensions"]]
                )

            super(CommentOptions, self).__init__(
                OrderedDict(
                    [
                        ("author", String(kwargs["author"])),
                        ("permlink", String(kwargs["permlink"])),
                        ("max_accepted_payout", Amount(kwargs["max_accepted_payout"])),
                        (
                            "percent_hive_dollars",
                            Uint16(int(kwargs["percent_hive_dollars"])),
                        ),
                        ("allow_votes", Bool(bool(kwargs["allow_votes"]))),
                        (
                            "allow_curation_rewards",
                            Bool(bool(kwargs["allow_curation_rewards"])),
                        ),
                        ("extensions", extensions),
                    ]
                )
            )


class AccountUpdate2(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            meta = ""
            if "json_metadata" in kwargs and kwargs["json_metadata"]:
                if isinstance(kwargs["json_metadata"], dict) or isinstance(
                    kwargs["json_metadata"], list
                ):
                    meta = json.dumps(kwargs["json_metadata"])
                else:
                    meta = kwargs["json_metadata"]

            posting_meta = ""
            if "posting_json_metadata" in kwargs and kwargs["posting_json_metadata"]:
                if isinstance(kwargs["posting_json_metadata"], dict) or isinstance(
                    kwargs["posting_json_metadata"], list
                ):
                    posting_meta = json.dumps(kwargs["posting_json_metadata"])
                else:
                    posting_meta = kwargs["posting_json_metadata"]

            super(AccountUpdate2, self).__init__(
                OrderedDict(
                    [
                        ("account", String(kwargs["account"])),
                        ("owner", Optional(None)),
                        ("active", Optional(None)),
                        ("posting", Optional(None)),
                        ("memo_key", Optional(None)),
                        ("json_metadata", String(meta)),
                        ("posting_json_metadata", String(posting_meta)),
                        ("extensions", Array([])),
                    ]
                )
            )


def isArgsThisClass(self, args):
    return len(args) == 1 and type(args[0]).__name__ == type(self).__name__
